/* Parka Client Library
 * Copyright (C) 1999 PLUS Group - University of Maryland
 *	
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * Authors: Jeff Heflin, Carolyn Gasarch, Brian Sullivan
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 *
 */

/* Change History:
   - 05/22/00, JDH: Removed use of kbLoader. Applications that wish to
                    open KBs in separate threads can manage it themselves
   - 05/19/00, JDH: Major reorginizational changes
   - 05/18/00, JDH: Use BufferedInput and BufferedOutput streams to read
                    and write from socket
   - 05/18/00, JDH: Use System.arraycopy() in getAnswersFromServer()
   - 05/18/00, JDH: Use numVars instead of calling getNumVariables in
                    getAnswersFromServer()
   - 05/02/00, JDH: Added the getKbNames() method
   - 03/02/00, JDH: Took care of case when the returned answers is
                    less than the number expected in simplyGetAnswers();
   */

package ParkaClient;

import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.Frame;  //for Cursor
import java.awt.Window; //for dialogs

/** A client interface for communicating with a Parka server. Provides
  methods to connect to a server, open a knowledge base, issue queries,
  and make updates.
 */
public class  ParkaNet  extends java.lang.Object
{
    // KB information
    protected  String    KBName;

  /** 'usingMem' refers to using memory vs. disk in internal
     Parka operations, ie, joins. */
    protected  boolean   usingMem;

    protected  String[]  predicates;
    protected  int[]     predArity;

    protected  ParkaWin  parent;

    protected  Socket       socket;
    protected  InputStream  instream;
    protected  OutputStream outstream;

    protected  boolean    kbOpen;
    protected  boolean    newPredList;

    // Hostname
    protected  String hostname = null;

    // Port number
    protected  int port = -1;

    // Query results
    protected  int    numTotalAnswersToQuery;
           //number of total tokens from the server is going to be--
           //(if all of the answers were requested at once, rather than
           // in batches):
           //numTotalAnswersToQuery * getNumVariables() --
           //                  ie, num Rows * num Cols
    protected  int    varStatus;
    protected  String     numAnswersMsg;
    protected  String[]   answers;
    protected  String[]   vars;

    static int   HAS_VARS = 1;
    static int   NO_VARS  = 2;
    static int   VARS_OTHER_STATUS  = 3;

    // Browser query results
    protected  String[][]  browserAnswers;
    protected  String      browserMsg;

    static final String  NOCONNMSG = "Not connected to server.";
    static final String  NOKBOPENMSG = "No KB open.";
    static final String  NOVARSINQUERY = "No variables retrieved";
       /* expected to match ResultsWin version.*/

  /* Constants added by JDH */
    static final int   FRAME_UNRESTRICT    = 1;
    //have the range already; get domain:
    static final int   FRAME_RESTRICT_IN   = 2;
    //have the domain already; get range:
    static final int   FRAME_RESTRICT_OUT  = 3;
    //have both domain and range:
    static final int   DATAVERIFY          = 7;
    // query types
    static  final int   PREDCONST  =  1;
    static  final int   FRAMECONST =  2;

    public String statusMsg;

    /*******
     Methods
     *******/

    /* we have a problem because want to display message to either an
       Applet or Frame. We can't subclass the Applet from an abstract
       class and using an interface won't tell us what object to
       store. Both are instanceof Container, but this doesn't help
    */
    public ParkaNet (ParkaWin parent)
    {
        // constructor
        this.parent = parent;
        socket = null;
	instream = null;
        outstream = null;

	initKBVars(true);
	initBrowserData();

	statusMsg = null;
    }


    public void clearNumAnswersToQueryMsg()
    {
	numAnswersMsg = null;
    }

    public void displayNumAnswersToQueryMsg()
    {
	if (numAnswersMsg != null)
	  parent.displayMessage (numAnswersMsg);
	//  parent.displayMessageKS (numAnswersMsg);
    }


  /** Make a socket connection to the Parka server. */
    public void connect()
    {
        // connect to Parka
        parent.displayMessage("Connecting to Parka");

	if (hostname == null || port == -1) {
	  parent.displayMessage("Conenct Failed: No host or port specified");
	  return;
	}

        try 
        {
            // close socket if one exists
            if (this.isConnection())
                this.disconnect();

            // Create new socket and streams
            socket = new Socket (hostname, port);
            instream = new BufferedInputStream(socket.getInputStream());
            outstream = new BufferedOutputStream(socket.getOutputStream());
            parent.displayMessage("Connection -- Waiting");
		// listening socket -- successful connect

	    // Determine if connection is accepted
	    Response  r = new Response (this);
	    r.computeResponse();
	    parent.displayMessage (r.getResponse());
	}
        catch (UnknownHostException uhe) 
        {
            parent.displayMessage("Connect Failed: Unknown Host Error");
        }
	/* BindException and ConnectException are java 1.1...
        catch (BindException be) 
        {
            parent.displayMessage("Connect Failed: " + be.getMessage());
	//TEST
	    System.out.println ("BindException");
        }
        catch (ConnectException ce) 
        {
            parent.displayMessage("Connect Failed: " + ce.getMessage());
	//TEST
            System.out.println ("Check if server is running " +
                           "or port is correct");
        }
	*/
        catch (SocketException se) 
        {
            parent.displayMessage("Connect Failed: Socket Failure");
	//TEST
	    System.out.println (se.getMessage());
        }
        catch (IOException ioe)
        {
            parent.displayMessage("Connect Failed: IO Error");
	//TEST
	    System.out.println (ioe.getMessage());
        }
	catch (IllegalArgumentException  iae) //possible with illegal port#
	{
            parent.displayMessage ("Connect Failed: " + iae.getMessage());
	//TEST
	    System.out.println (iae.getMessage());
	}
    }


  /** Disconnect from the Parka server. Close the knowledge base if it is
    still open. */
    public void disconnect()
    {
        // disconnect from Parka if connected
        if (isConnection())
        {
            // close open KB, if any
            if (isKBOpen())
	        closeKB();

            // now disconnect
            parent.displayMessage("Disconnecting from Parka");
            try
            {
                // Tell parka server we are exiting;
		// parka doesn't give a response
		Command c = new Command (this, Command.DISCONNECT, "");
                c.sendCommand();

                // Close streams
                instream.close();
                outstream.close();
                socket.close();
            
                instream = null;
                outstream = null;
                socket = null;
            }
            catch (IOException e1)
            {
                // Close Error
                System.out.println("IO Error in disconnect()");
            }
        }
    }


  /** Returns true if there is a connection to PARKA open. */
    public boolean isConnection()
    {
        return (socket != null);
    }


  /** Change hostname and port used to connect to PARKA */
    public void changeParkaHost(String newhost, int newport)
    {
        hostname = newhost;
        port = newport;
        
        // change PARKA connection
        this.disconnect();
        this.connect();
    }


  /** Returns true if a knowledge base is currently open on the connection. */
    public synchronized boolean isKBOpen()
    {
	return   kbOpen;
    }


  /** Open the KB with the given name. */
  /* Note: This routine may be used from within a separate thread; so
     we need to be careful about synchronization and interference. */
    public void openKB (String newkbName, boolean memoryp)
    {
	// System.out.println ("Entered 'openKB()'");

	if (isConnection())
        {
	    // System.out.println ("presence of connection is established.");

	    // not set from any other place:
	    KBName = newkbName;
	    usingMem = memoryp;
	    /* System.out.println ("   KBname = [" + KBName +
                        "] ; usingMem = " + usingMem);
            */
            // load a specific database
            parent.displayMessage("Loading " + KBName);

            Command  c = new Command (this, Command.OPENKB, KBName);
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();
	    // System.out.println ("   returned status is: " + r.getStatus());

            if (r.getStatus() == 0)
            { 
                setKBStatus (true);
		parent.displayMessage("Finished loading " + KBName);
            }
	    /*
	    // Ignore the -2 status until Merwyn fixes Parka 
	    else if (r.getStatus() == -2) {
	      setKBStatus(true);
	      parent.displayMessage("Received false 'KB in use' message. Finished loading " + KBName);
	    }
	    */
	    else
	      parent.displayMessage (r.getResponse());
        }
    }

  /** Close the current knowledge base. */
    public void closeKB()
    {
        if (isKBOpen())
        {
	    if (isConnection())
            {
                // Close the current KB
                parent.displayMessage("Closing KB");

                Command  c = new Command (this, Command.CLOSEKB, "");
                c.sendCommand();
                Response r = new Response (this);
	        r.computeResponse();

		if (r.getStatus() >= 0)  // succeeds
	        {
		    //reinit when closed
    		    initKBVars(false);
		    initBrowserData();
		    /* used on for Parka Interface tool */
		    /*
		    parent.reinitWindow();
		    parent.reinitBrowser();
		    */
		}

                parent.displayMessage (r.getResponse());
            }
	    else
                parent.displayMessage (NOCONNMSG);
        }
	else
            parent.displayMessage ("No KB is open");
    } 


  /** Given a string representing a Parka query, returns a QueryResults
    object that contains the answers, or null on error. The query string
    should be the standard Parka format:
    (pred#arg1#arg2...)(pred#arg1#arg2...)....
    */
    public QueryResults query (String queryString)
    {
//      System.out.println("Entering query(\"" + queryString + "\")");
	statusMsg = null;            // Reset status

	if (!isConnection())
        {
	    statusMsg = NOCONNMSG;
	    return  null;
	}
	if (!isKBOpen())
	{
	    statusMsg = NOKBOPENMSG;
	    return  null;
	}

	if ((queryString == null) || (queryString.equals("")))
	{
	    statusMsg = "Empty query.";
	    return  null;
	}

	// System.out.println("Sending query command");
	Command c = new Command (this, Command.QUERY, queryString);
        c.sendCommand();

        Response r = new Response (this);
        r.computeResponse();

	numTotalAnswersToQuery = r.getStatus();
	// System.out.println("Got " + numTotalAnswersToQuery + " answers");

        if (numTotalAnswersToQuery < 0)       //then really have an error
        {
	    statusMsg = r.getResponse();
	    return  null;
        }
	else    //if (numTotalAnswersToQuery >= 0)
	{
	    String[] varList = simplyGetVariables();
/* TEST
	    System.out.println ("varList.length == " + varList.length);
	    for (int v = 0; v < varList.length; v++)
	        System.out.println ("   " + varList[v]);
*/
	    if (varStatus == NO_VARS)
	    {
                numAnswersMsg = " Answer is TRUE";
                statusMsg = "Done. " + numAnswersMsg;
                return  null;  // constant expression found
	    }
	    else    //varStatus == HAS_VARS
            {
                numAnswersMsg = " " + numTotalAnswersToQuery
                                       + " answers.";
                statusMsg = numAnswersMsg;

	        QueryResults qr = new QueryResults (
                                 numTotalAnswersToQuery, varList.length);
	        qr.setNames (varList);
                qr.setValueMatrix (simplyGetAnswers (qr.numResults(),
			 		    qr.numNames()));
		return  qr;
            }
        }
    }


  /** Given a Parka query string, sends the query to the server and
    determines the status of this query. Messages are returned using
    the displayMessage() interface. Calls getVariables() but, in order
    to get the answers, the caller must also call getAnswers(). 
    */
    public void submitQuery (String query)
    {
	if (!isConnection())
        {
	    parent.displayMessage(NOCONNMSG);
	    return;
	}

	if (!isKBOpen())
	{
	    parent.displayMessage("No open KB");
	    return;
	}

        parent.displayMessage("Submitting Query");

        // submit a query to parka engine
	if (query.equals(""))
	{
            parent.displayMessage ("No query formed");
	    return;
	}

        Command  c = new Command (this, Command.QUERY, query);
        c.sendCommand();
        Response r = new Response (this);
        r.computeResponse();

	numTotalAnswersToQuery = r.getStatus();

        if (numTotalAnswersToQuery < 0)   //then really have an error
        {
/*  TEST --
    what prints if we let 'r.getResponse()' print??

	    if (numTotalAnswersToQuery != Response.XCEPTION)
	    {
		// either boolean answer of "False", or no way to match
		// the query graph to any data in the database.  Could
		// include use of constants not in the database, or in
		// the database in a different context (eg. a predicate
		// used in a domain or range position)
	        //
                parent.displayMessage ("No results found");
	    }
	    else
*/
	        parent.displayMessage (r.getResponse());

            answers = null;
            vars = null;
        }
        else    //if (numTotalAnswersToQuery >= 0) 
        {
	  /* Need to get variables in order to determine if the
	     response should be a TRUE/FALSE or number of answers */
            vars = getVariables();

	    if (varStatus == NO_VARS)
	    {
		numAnswersMsg = " Answer is TRUE";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
	    else   //varStatus == HAS_VARS
	    {
		numAnswersMsg = " " + numTotalAnswersToQuery
                                       + " answers.";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
        }
    }


  /** Submit a "browsing" style query. These queries have a specialized
    expected structure -- see comments on server side for more info.
    Note that browse queries can only be issued on binary KBs.
    Returns contents of an error message, if any.
    */
    public int  submitBrowsingQuery (String  query)
    {
	int  status;
	browserMsg = "";
/*
	status = -4;
	System.out.println ("Browser not available for this version");
*/
        Command  c = new Command (this, Command.JBROWSERQUERY, query);
        c.sendCommand();
        BrowserResponse r = new BrowserResponse (this);
        r.computeResponse();
	status = r.getStatus();

	if (status >= -1)   // 0 is the highest possible
	{
	    r.makeTokens();
	    browserAnswers = r.getBrowserTokenList();
	    if (status == -1)   // sample of all the data only
	    {
	        Vector  columnType = r.getColType();
	        Vector  numberAns = r.getNumAnswers();
		Vector  numberShown = r.getNumAnswersShown();

		/* below line is specific to the Browser Window interface */
		// printBrowserSampleMsgs (columnType, numberAns, numberShown);
	    }
	}
	if (status == -2)  //corrs to JB_INCOMPLETEDATA; get more data:
	{
	    r.makeTokens();

	    while (status == -2)  // 0 is a successful end
	    {
		c = new Command (this, Command.JBRWQUERYCONT, "");
		c.sendCommand();
		r.computeResponse();
		status = r.getStatus();

		if (status >= -2)
		{
	            r.makeMoreTokens();
		}
	    }
	    if (status >= -1)
	    {
	        browserAnswers = r.getBrowserTokenList();
	        if (status == -1)
		{
	            Vector  columnType = r.getColType();
	            Vector  numberAns = r.getNumAnswers();
		    Vector  numberShown = r.getNumAnswersShown();
	 /* below line is specific to the Browser Window interface */
	            // printBrowserSampleMsgs (columnType, numberAns, numberShown);
		}
	    }
	    else
	        browserMsg = r.getResponse();
	}
	else  if (status == -3)
	    browserMsg = "No more data received; server out of memory";
	else  if (status == -4)
	    browserMsg = r.getResponse();
	return  status;
    }

  /** Gets start...end answers from the server and puts them in
    the 'answers' array. Here, an answer is one complete set of
    bindings for the variables. This routine should only be called after
    a query has previously been sent and getVariables() has been
    called. This routine will return the
    number of answers requested by the user, even if it requires
    making multiple calls to the Parka server to do so. */
    public String[] getAnswers(int start, int end)
    {
      int numDesired;            // number of answers to retrieve from server
      int numGotten = 0;         // number retrieved so far
      int numVars = 0;           // number of variables in query
      int leftToGet;             // number left to retrieve
      int nextIndex;             // index of next desired answer
      String[] curAnswers;
      Command c;
      Response r;

      // System.out.println("getAnswersFromServer(" + start + "," + end + ")");

      if (isConnection())
        {
	  //Nothing to process here:
	  if (numTotalAnswersToQuery <= 0)
	    return new String[0];

	  if ((start < 1) || (end < 1))
            {
	      System.out.println ("ParkaNet.getAnswers(): 'start' 'and' " +
				  "end parameters must be greater than zero");
	      return new String[0];
            }
	  if (end < start)
            {
		System.out.println ("ParkaNet.getAnswers(): 'end' parameter " +
				    "cannot be less than 'start'");
		return new String[0];
            }

	  /* the Parka server may not return the requested number of
	     answers if its buffer fills up. In that case, we must
	     continue requesting answers until we get all that we
	     want. */
	  numDesired = end - start + 1;
	  numVars = getNumVariables();
	  // need space for each variable of each answer
	  answers = new String[numDesired * numVars];
	  numGotten = 0;
	  leftToGet = numDesired;
	  nextIndex = start;

	  while (leftToGet > 0) {
	    /* note: according to parka server docs, the second parameter
	       of GETBINDINGS should be the number of desired answers,
	       not the index of the last ones desired. However, we
	       need to test to see if this is truly the case. */
	    c = new Command (this, Command.GETBINDINGS,
			     nextIndex + " " + leftToGet);
	    c.sendCommand();
	    r = new Response (this);
	    r.computeResponse();
	    if (r.getStatus() <= 0) {
	      System.out.println("ParkaNet.getAnswers(): Error in " +
				 "retrieving bindings.");
	      return new String[0];
	    }
	    r.makeTokens();
	    curAnswers = r.getTokenList();

	    /* copy current set of answers into total answers array. */
	    //for (int i=0; i < curAnswers.length; i++)
	    //  answers[(numGotten * numVars) + i] = curAnswers[i];
	    System.arraycopy(curAnswers, 0, answers, numGotten * numVars,
			     curAnswers.length);

	    /* the following three variables treat an answer as a 
	       complete set of bindings to the variables. However,
	       the server returns answers as an array of bindings
	       to single variables. */
	    numGotten += r.getStatus();
	    leftToGet -= r.getStatus();
	    nextIndex += r.getStatus();

	    /*
	    if (leftToGet > 0) {
	      System.out.println("In ParkaNet.getAnswers():");
	      System.out.println("   numDesired=" + numDesired + "\n" +
				 "   numGotten=" + numGotten + "\n" +
				 "   nextIndex=" + nextIndex + "\n" +
				 "   leftToGet=" + leftToGet + "\n" +
				 "   curAnswers.length=" + curAnswers.length);
	    }
	    */
	  }
/* TEST
	    if (answers == null)
	        System.out.println ("'answers' is null");
	    else
	        System.out.println ("'answers' has " + answers.length
                                             + " answers.");
*/
        }
      else
	{
	  parent.displayMessage (NOCONNMSG);
	}
      return answers;
    }

  /** Returns an array of answers that have been requested in response
    to a previously issued query. */
  /*
    public String[] getAnswers()
    {
        return   answers;
    }
    */

  /** Returns a 2-D array of the answers to the browsing query. */
    public String[][] getBrowserAnswers()
    {
        return   browserAnswers;
    }

  /** Retrieves the list of variables in the order that values are
    bound to them. */
    public String[] getVariables()
    {
	if (isConnection())
        {
            Command  c = new Command (this, Command.GETVARIABLES, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();

	    if (r.getStatus() > 0)
	    {
	        r.makeTokens();
	        vars = r.getTokenList();
	        varStatus = HAS_VARS;
	    }
	    /* can't use QueryResults in the general case */
	    /*
	    else  if ( (getNumTotalAnswersToQuery() == 0) &&
                       ((r.getResponse()).equals(QueryResults.NOVARSINQUERY)) )
                                 // [r.getStatus() == -1, in this case]
	     */
	    else  if (getNumTotalAnswersToQuery() == 0)
	    {
	        varStatus = NO_VARS;
		vars = null;
	    }
	    else
	    {
	        varStatus = VARS_OTHER_STATUS;
		vars = null;
	    }
        }
	else
        {
            parent.displayMessage (NOCONNMSG);
            vars = null;
        }

        return vars;
    }

  /** Returns a list of KBs that are available to the connection. */
  public String[] getKbNames()
  {
    parent.displayMessage("Retrieving KB list");
    String[]  kblist;

    // Returns a list of available Knowledge Bases
    if (isConnection())
      {
	Command  c = new Command (this, Command.GETKBNAMES, "");
	c.sendCommand();
	Response r = new Response (this);
	r.computeResponse();
	if (r.getStatus() > 0)
	  {
	    r.makeTokens();
	    kblist = r.getTokenList();
	    statusMsg = "KB names retrieved";
	  }
	else
	  {
	    statusMsg = r.getResponse();
	    kblist = new String[0];
	    parent.displayMessage (r.getResponse());
	  }
      }
    else
      {
	parent.displayMessage (ParkaNet.NOCONNMSG);
	kblist = new String[0];
      }
    
    return   kblist;
  }


  /** Get the size stats for the currently loaded KB. */
    public String[] getKbStats()
    {
        String[] kbStats = null;

	if (isConnection())
        {
            Command  c = new Command (this, Command.PARKACONFIG, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();
	    r.makeTokens();
	    kbStats = r.getTokenList();
	    // no display done
        }
	else
	    parent.displayMessage (NOCONNMSG);

	return kbStats;
    }

  /** Get the timing statistics for the last query. Returns the
   times as an array of strings. */
    public String[] getTimingStats()
    {
	String[] timeStats = null;

	if (isConnection())
        {
            Command  c = new Command (this, Command.GETTIMINGS, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();
	    r.makeTokens();

	    timeStats = r.getTokenList();
	    // no display done	    
        }
	else
	    parent.displayMessage (NOCONNMSG);

	return timeStats;
    }

  /** Gets a list of predicates relevant to a particular database.
    User class is expected to call this routine after the thread
    loading the requested KB has successfully completed. */
    public String[] getPreds()
    {
	String[]  predList = null;

        // Get predicate list from PARKA
	if (isConnection())
        {
	    if (isKBOpen())
	    {
	      // using the "KS" version of displayMessage so that the
	      // active Node is not disturbed (When the link/predicate
	      // is created, we need to know this to assign the "from"
	      // node.
	      // parent.displayMessageKS("Retrieving Predicates");
	      parent.displayMessage("Retrieving Predicates");
	      
	      boolean  morePredsToGet = true;
	      boolean  inLoopAgain = false;
	      while (morePredsToGet)
		{
		  String[] tmp, tmp2;
		  
		  Command  c = new Command (this, Command.GETPREDICATES, "");
		  c.sendCommand();
		  Response r = new Response (this);
		  r.computeResponse();
		  
		  // status == 0 if have no more preds to get.
		  if (r.getStatus() >= 0)
		    {
		      // System.out.println ("get preds:  status > or = 0");
		      r.makeTokens();
		      tmp = r.getTokenList();
		    }
		  else //r.getStatus() < 0: error occurred while getting preds
		    {
		      // System.out.println ("get preds:  status < 0");
		      predList = null;
		      parent.displayMessage (r.getResponse());
		      break;
		    }
		  
		  /* append new pred list to end of the old one */
		  morePredsToGet = (r.getStatus() > 0);
		  if (inLoopAgain)
		    {
		      // System.out.println ("  get preds:  " +
		      //              "Getting preds mult times thru loop");
		      tmp2 = predList;
		      predList = new String[tmp.length + tmp2.length];
		      // keep in original order:
		      System.arraycopy (tmp2, 0,
					predList, 0, tmp2.length);
		      System.arraycopy (tmp, 0,
					predList, tmp2.length, tmp.length);
		    }
		  else
		    predList = tmp;

		  inLoopAgain = true;
		}
                // if r.getStatus() == 0, we have all the preds, and
                // everything is ok.
	    }
	    else
	      {
		parent.displayMessage("No KB open");
	        predList = null;  //assure null value when no connection
	    }
        }
	else
	{
	    parent.displayMessage (ParkaNet.NOCONNMSG);
	    predList = null;  //assure null value when no connection
	}
/* TEST
        if (predList == null)
            System.out.println ("Predicates are null");
        else
            System.out.println (predList.length + " Predicates");
*/
        return    predList;
    }


    /* When we know that valid commands have been responded to, without
       error for the commands GETPREDICATES and GETVARIABLES, then it
       is valid to call the next two routines.    */

  /** Retrieves the list of predicates and their arities from the KB
    and stores them for later use. Returns the predicate array */
    public String[] getPredicateList()
    {
	if (isKBOpen() && newPredList)
        {
	    predicates = getPreds();
	    newPredList = false;  //if predicates is still null,
                   // something is wrong.

	    if (predicates != null)
	    {
	        predArity = getPredsArity();
		     // need arity for each predicate, but client's user
		     // doesn't need to know this call is needed.
		     // its done 'automatically' from user's view.

	        if (predicates.length != predArity.length)
		    System.out.println ("Programming error: list of predicates"
                       + " and list of pred arities are not equal length");
	    }
        }
	else  if (!isKBOpen())
	    parent.displayMessage ("No KB is open");

        return predicates;
    }


  /** Returns the arity of the predicate with the given name. */
  /* Note: If a new predicate is added since the last call to
     getPredicateList(), this will be out of date. */
  /* should return 'null' if 'getPredicateList()' has not been called first.
   */
    public int getPredicateArity (String predname)
    {
	if (predicates == null)
	    return  0;

	int i;
	for (i = 0; i < predicates.length; i++)
	    if (predname.equals (predicates[i]))
		break;

	if (i > predArity.length)
	    return  0;
	else
	    return predArity[i];
    }

    public String[] getVariableList()
    {
        return vars;
    }

    //same as number of columns for output display of query results
    //
    public int getNumVariables()
    {
	if (vars == null)
	    return   0;
	else
	    return   vars.length;
    }

  /** Returns the number of answers (complete sets of variables bindings)
    generated by the last query sent to the server. In relational
    database terms, this is the number of rows returned.*/
    public int getNumTotalAnswersToQuery()
    {
	return  numTotalAnswersToQuery;
    }


    /* added by Heflin on 6/7/98 */
  /** Returns all the direct descendant categories of a particular
    category. This method has been superceded by the getDescendants()
    method. */
    public String[] getImmediateDescendants(String parent) {

      String query;
      String[][] answers;
      int status;

      // System.out.println("Getting descendants of " + parent);
      query = String.valueOf(FRAME_RESTRICT_IN) + " " +
        String.valueOf(PREDCONST) + " isa\n" +
        String.valueOf(FRAMECONST) + parent + "\n";

      /*
      Command c = new Command(this, Command.GETJBROWSERQUERY, query);
      c.sendCommand();
      // System.out.println("Sent command");
      BrowserResponse r = new BrowserResponse(this);
      // System.out.println("Computing response");
      r.computeResponse();
      // System.out.println("Computed response");
      r.makeTokens();

      // might want to handle when more answers than can be returned

      // System.out.println("Made tokens");
      answers = r.getBrowserTokenList();
      */

      status = submitBrowsingQuery(query);
      answers = getBrowserAnswers();
      if (answers != null)
        return answers[0];
      else
        return null;
    }

    /** Given the name of a parent category gets all sub-categories of it.
      If justCats is false, then it also returns the instances. In both cases,
      this returns an array of triplets of the form (parent, child-type, child)
      where child type is "s" if it is a sub-category and "i" if it is an
      instance. This function is particularly useful in a client server
      environment because issuing a separate query for each category can be 
      much slower
    */
    public String[][] getDescendants(String parent, boolean justCats) {

      Command c;
      String query;
      String[] answers;
      String[][] pairs;
      int status;
      int rows;
      int i;

      System.out.println("Getting descendants of " + parent);

      if (justCats)
        c = new Command(this, Command.GETDESCPAIRS, "s " + parent);
      else
        c = new Command(this, Command.GETDESCPAIRS, "a " + parent);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      r.makeTokens();

      // might want to handle when more answers than can be returned
      answers = r.getTokenList();
      rows = answers.length / 3;
      // System.out.println("Num Answers = " + answers.length + "\tRows = " + rows);
      pairs = new String[rows][3];
      for (i = 0; i < rows; i++) {
        pairs[i][0] = answers[i * 3];
        pairs[i][1] = answers[i * 3 + 1];
        pairs[i][2] = answers[i * 3 + 2];
        // System.out.println(pairs[i][0] + "\t" + pairs[i][1] + "\t" + pairs[i][2]);
      }

      return pairs;
    }

    /** A version of getDescendants that automatically returns instances
       as well as categories.
    */
    public String[][] getDescendants(String parent) {

      return getDescendants(parent, false);
    }

    /* Added by Heflin 7/2/98
    */
    public void insertToKB(String assertion) {

      Command c = new Command(this, Command.INSERTTOKB, assertion);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      if (r.getStatus() < 0) 
	statusMsg = r.getResponse();
      else 
	statusMsg = "Sucessful insert";
      // parent.displayMessage(statusMsg);
    }  

  /** Given an assertion (in standard Parka format, including the
    enclosing parentheses), removes it from the KB. If
    there are multiple identical assertions, all of them are
    removed.
    */
    public void deleteFromKB(String assertion) {

      Command c = new Command(this, Command.DELETEFROMKB, assertion);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      if (r.getStatus() < 0) 
	statusMsg = r.getResponse();
      else 
	statusMsg = "Sucessful deletion";
      // parent.displayMessage(statusMsg);
    }  


  /** Inserts a new frame into the KB. */
    public void insertFrame(String frame) {

      Command c = new Command(this, Command.INSERTFRAME, frame);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      if (r.getStatus() < 0) 
	statusMsg = r.getResponse();
      else 
	statusMsg = "Sucessful insert";
      // parent.displayMessage(statusMsg);
    }

  
  /** Deletes a frame from the KB. */
    public void deleteFrame(String frame) {

      Command c = new Command(this, Command.DELETEFRAME, frame);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      if (r.getStatus() < 0) 
	statusMsg = r.getResponse();
      else 
	statusMsg = "Sucessful deletion";
      // parent.displayMessage(statusMsg);
    }

  /** Inserts a predicate of the given arity into the KB. Currently,
    insertions of predicates with the same name as an existing predicate
    are ignored by Parka. Calling this method on a binary KB may
    result in an error. */
    public void insertPredicate(String predname, int arity) {

      String data;

      data = predname + " " + arity + " 0";
      Command c = new Command(this, Command.INSERTPREDTOKB, data);
      c.sendCommand();
      Response r = new Response(this);
      r.computeResponse();
      if (r.getStatus() < 0) 
	statusMsg = r.getResponse();
      else 
	statusMsg = "Sucessful insert of predicate";
      // parent.displayMessage(statusMsg);
    }



  /* ******** Package methods *********** */

  /** Returns the class to which messages are sent. */
    ParkaWin  getWin()
    {
	return   parent;
    }


  /** Returns the socket input stream. */
    InputStream  getInStream()
    {
        return  instream;
    }


  /** Returns the socket output stream. */
    OutputStream  getOutStream()
    {
        return  outstream;
    }


  /** Returns the latest message from a browser query. */
    String getBrowserMsg()
    {
	return  browserMsg;
    }


  /* ******** Private Methods *********** */

    //some of these are also reinitialized with a new query
    private void initKBVars (boolean inConstruct)
    {
        setKBStatus (false);
        KBName = null;
	usingMem = true;   //may be reset later by chosen KB. Default.
        numTotalAnswersToQuery = 0;
	numAnswersMsg = null;
        predicates = null;
	predArity = null;
        answers = null;
        vars = null;
	varStatus = VARS_OTHER_STATUS;
	newPredList = true;   // 'predicates' not init'd yet

	/*
	if (!inConstruct)
            ((QueryWin)parent).updateFont();
	*/
    }

    private void  initBrowserData()
    {
        browserAnswers = null;
	browserMsg = "";
    }


  /** Set the open/close status of the KB. */
    private synchronized void setKBStatus (boolean status)
    {
	kbOpen = status;
    }


  /** Returns an array of variables in the order in which values have
    been bound to them in the last query issued. */
  /* NOTE: This function is almost identical to getVariables(), it
     only differs in its criteria for determining when to display
     a No Variable message. */
  /* Consider integrating this with getVariables() */
    private String[] simplyGetVariables()
    {
	if (!isConnection())
        {
	    statusMsg = NOCONNMSG;
	    return null;
	}
	if (!isKBOpen())
	{
	    statusMsg = NOKBOPENMSG;
	    return null;
	}

	Command  c = new Command (this, Command.GETVARIABLES, "");
	c.sendCommand();
	Response r = new Response (this);
	r.computeResponse();

	if (r.getStatus() > 0)
	{
	    varStatus = HAS_VARS;
	}
        else  if ((getNumTotalAnswersToQuery() == 0) &&
                       (r.getResponse()).equals (NOVARSINQUERY))
                                 // [r.getStatus() == -1, in this case]
 	{
	    varStatus = NO_VARS;
	    return null;
	}
	else
	{
	    varStatus = VARS_OTHER_STATUS;
	    return null;
	}

	r.makeTokens();
	return  (r.getTokenList());
    }

  /* Consider integating this with getAnswers() */
  private String[][] simplyGetAnswers (int numResults, int numNames)
    {
	String[]   ansList;
	String[][] ansMat = new String[numResults][numNames];
	int        expectedAnswers;       // number of expected answers

	expectedAnswers = numResults * numNames;

	if (!isConnection())
        {
	    return null;
	}

	Command  c = new Command (this, Command.GETBINDINGS,
				  1 + " " + expectedAnswers);
	c.sendCommand();
	Response r = new Response (this);
	r.computeResponse();

	// if (r.getStatus() <= 0)
	if (r.getStatus() < 0)
	{
	    System.err.println ("Error in ParkaNet.simplyGetAnswers(): " +
				"Server returned status code of " +
				 r.getStatus());
	    return null;
	}

	r.makeTokens();
	ansList = r.getTokenList();
/* TEST
	System.out.println ("Answer List:");
	for (int x = 0; x < ansList.length; x++)
	{
	    if ((x > 0) && ((x % 8) == 0))
		System.out.println("");
	    System.out.print (" <" + ansList[x] + "> ");
	}
*/
	if (ansList.length > expectedAnswers) {
	  System.err.println("Error in Parka.simplyGetAnswers(): " +
			     "Received more answers than expected.");
	}
	if (ansList.length < expectedAnswers)
	{
	    /* It turns out that the most likely reason for this is that the
               client asked for a chunk of data too large to fit into the
               server's buffer.  If the server's buffer fills up, it just
               quits, and tells us how many rows it was able to send, out of
               the number we asked for.  So we either need to accept what we
               have, or ask for the rest;  we are going to keep asking until
               we get the total number we wanted.
            */

	    if (r.getStatus() < 0)  //error
		return  null;
/* TEST
	    System.out.println ("   Answer list length:  " + ansList.length);
	    System.out.println ("   Answer list length:  " + ansList.length);
	    System.out.println ("   Number of rows:  " + numResults);
	    System.out.println ("   Number of columns:  " + numNames);
*/
	    int  numGotten = r.getStatus();
	    int  leftToGet = numResults - numGotten;

	    // Convert answer list into value matrix...
	    makeValueMatrix (0, numGotten, numNames, ansList, ansMat);

	    while (leftToGet > 0)
	    {
        	c = new Command (this, Command.GETBINDINGS,
                                (numGotten +1) + " " + leftToGet * numNames);
			/* we'll ask for this much; may not get it all ... */
        	c.sendCommand();
        	r = new Response (this);
        	r.computeResponse();
		if (r.getStatus() > 0)
	            r.makeTokens();
		else
		{
		    System.out.println ("Unexpected error retreiving answers.");
		    System.out.println ("   Status == " + r.getStatus());
		    return  null;
		}
        	ansList = r.getTokenList();

	        makeValueMatrix (numGotten, r.getStatus(),
                                    numNames, ansList, ansMat);

		leftToGet -= r.getStatus();
		numGotten += r.getStatus();
	    }
	}
	else
	{
	    // Convert answer list into value matrix...
	    makeValueMatrix (0, numResults, numNames, ansList, ansMat);
	}
	return  ansMat;
    }

    private void makeValueMatrix (int rowStart, int numRows, int numCols,
                                  String[] ansList, String[][] ansMat)
    {
/* TEST
	System.out.println ("VALUE MATRIX:  rowStart == " + rowStart +
                     " ; last row == " + (rowStart + numRows -1) +
                           " ; numCols == " + numCols);
*/ 
	for (int k = 0, i = rowStart; i < rowStart + numRows; i++, k++)
	{
	    for (int j = 0; j < numCols; j++)
	    {
		ansMat[i][j] = ansList[k * numCols + j];
	    }
	}
    }

    // should not be called if server call GETPREDICATES failed.
    //
    private int[] getPredsArity()
    {
      String[] arityTokens;
      int[]    arityList = null;
      int      i;
      
      Command c = new Command (this, Command.GETPREDARITYS, "");
      c.sendCommand();
      // System.out.println ("In 'getPredsArity()'");
      
      Response r = new Response (this);
      r.computeResponse();
      
      if (r.getStatus() == 0)
	{
	  r.makeTokens();
	  arityTokens = r.getTokenList();
	  arityList = new int[arityTokens.length];
	  for (i=0; i < arityList.length; i++)
	    arityList[i] =  Integer.parseInt(arityTokens[i]);	 
	}
      else
	parent.displayMessage (r.getResponse());

      return  arityList;
    }



  /*  ************ OLD METHODS ***************** */

  /* Methods that were originally used by the Parka Interface are kept
     here for archival purposes and to server as a reminder of what
     needs to be done to get the Parka Interface code to use the
     ParkaClient library */

  /*
    // Create PortDialog so user can change hostname and portnumber
    public void doPortDialog()
    {
        PortDialog pdialog = new PortDialog (parent, this, hostname, port);
	if (ApplType.getIsVersion1_0())
            ((Window)pdialog).show(); //Netscape thinks we want Dialog.show()
             //which doesn't exist until java 1.1
	else
            pdialog.show();
    }
    */



  /* Note: since this calls loadKB(), and loadKB creates a KBDialog,
     this function tightly couples interaction with the KB to presentation */
  /*
    // initialize connection to a new KB; uses a dialog box
    // eventually 'initLoadKBbyName()' gets called, once the KB
    // to load is known.
    //
    public void initKBLoading()
    {
	if (isConnection())  //assures the 2 streams already exist
	{
	    kbLoader = new NetKBInit (this, instream, outstream);
		//this class requires in/out streams to already be set up.
	    kbLoader.loadKB();
	}
	else
            parent.displayMessage (NOCONNMSG);
    }
   */

/*
    // 'links' are the predicates in the query graph
    public void submitQuery (Vector links)
    {
	if (!isConnection())
        {
	    parent.displayMessage(NOCONNMSG);
	    return;
	}
	if (!isKBOpen())
	{
	    parent.displayMessage("No open KB");
	    return;
	}

        kbLoader.join();
        parent.displayMessage("Submitting Query");

        // submit a query to parka engine
        StringBuffer querystring = new StringBuffer();

        //  gather query info
        Enumeration preds = links.elements();
        while (preds.hasMoreElements())
        {
            Link currlink = (Link)(preds.nextElement());
            querystring.append ("(" + currlink.getObjName());

	    Vector linkArgs = currlink.getLinkArguments();

	    for (int i = 0; i < linkArgs.size(); i++)
	    {
                querystring.append ("#");
		querystring.append ((String) linkArgs.elementAt(i));
	    }
            querystring.append (")");
        }

        // error checking
	String   finalQuery = new String (querystring);
	if (finalQuery.equals(""))
	{
            parent.displayMessage ("No query formed");
	    return;
	}

        Command  c = new Command (this, Command.QUERY, finalQuery);
        c.sendCommand();
        Response r = new Response (this);
        r.computeResponse();

	numTotalAnswersToQuery = r.getStatus();

        if (numTotalAnswersToQuery < 0)   //then really have an error
        {
*/
/*  TEST --
    what prints if we let 'r.getResponse()' print??

	    if (numTotalAnswersToQuery != Response.XCEPTION)
	    {
		// either boolean answer of "False", or no way to match
		// the query graph to any data in the database.  Could
		// include use of constants not in the database, or in
		// the database in a different context (eg. a predicate
		// used in a domain or range position)
	        //
                parent.displayMessage ("No results found");
	    }
	    else
*/
/*              parent.displayMessage (r.getResponse());

            answers = null;
            vars = null;
        }
        else //if (numTotalAnswersToQuery >= 0) 
        {
            vars = getVariables();

	    if (varStatus == NO_VARS)
	    {
		numAnswersMsg = " Answer is TRUE";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
	    else   //varStatus == HAS_VARS
	    {
		numAnswersMsg = " " + numTotalAnswersToQuery
                                       + " answers.";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
        }
    }
*/


/*
    void printBrowserSampleMsgs (Vector coltype, Vector numans,
                                 Vector numshown)
    {
	int  vectsize = coltype.size();
	if (vectsize != numans.size())
	{
	    System.out.println ("Programmer Error: column type vector "
                 + "and number of answers in each column vector are "
                 + "not the same size.");
	    return;
	}

	StringBuffer msgBuf = new StringBuffer (
                     "Too much data; only a sample is used.");

	for (int i = 0; i < vectsize; i++)
	{
	    if ((i % 2) == 1)
		msgBuf.append ("\nColumn ");
	    else
	        msgBuf.append ("  Column ");
	    String  col = BrowserWin.getColumnName (
                                               (String)coltype.elementAt(i));
	    String  num = (String)numans.elementAt(i);
	    String  shown = (String)numshown.elementAt(i);
	    msgBuf.append (col);
	    msgBuf.append (" has ");
	    msgBuf.append (num);
	    msgBuf.append (" total answers. ");
	    msgBuf.append (shown);
	    msgBuf.append (" shown.");
	}
	browserMsg = new String (msgBuf);
    }
*/

}
