/* Copyright (C) 1999 PLUS Group - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ParkaNet.java
 * and the file License.html for details.
 */

package ParkaClient;

import   java.util.*;

/** A batch of query results, when results are more than just 'true'
 *  or 'false'. 
 */
public class QueryResults
{
    private String[]   names;               // There are M names (col titles)
    private Vector     results;             // This is an Vector (size N) of
                                            // size M value arrays
   
  /** Construct a new result set with n rows and m columns. */
    public QueryResults(int n, int m)
    {
	names = new String[m];       //num columns
	results = new Vector(n);     //num answers
    }

    //not used currently
    public QueryResults (boolean answer)
    {
	// And to preserve all the methods using lengths of the arrays...
	names = new String[0];
	results = new Vector(0);
    }




  /** Returns the number of results that will fit in the object.
    This is odd, but the prior reasoning was: 

    "Use the capacity, not size of the vector, since we need to be
    able to know how many answers will fit.  Make sure that we keep
    the capacity trimmed to reflect the correct number of answers
    if any are removed..."

    May consider changing this. Will have to see how it is used
    in code.
    */
    public int numResults()
    {
	return results.capacity();
    }


  /** Returns the number of variables (columns) used in the
    result. */
    public int numNames()
    {
	return names.length;
    }

  
  /** Returns the name of the variable in the pos column (starting
    with 0). */
    public String getName(int pos)
    {
	return names[pos];
    }

  
  /** Return the names of all variables in the result. */
    public String[] getNames()
    {
	return names;
    }


  /** Return the value at a particular row and column number. */
    public String getValue (int row, int col)
    {
	QueryResult qr = (QueryResult) results.elementAt(row);
		//helper class; (see def below)
	String     val;

	try
	{
	    val = qr.valArray[col];
	}
	catch (ArrayIndexOutOfBoundsException  exc)
	{
	    System.out.println ("Failed index: " + exc.getLocalizedMessage());
	    exc.printStackTrace();
	    val = "";
	}

	return  val;
    }

  /** Return the set of variable bindings for a single answer (a row
    from the table). */
    public String[] getValues(int pos)
    {
	QueryResult qr = (QueryResult) results.elementAt(pos);
	return qr.valArray;
    }


  /** Return the results as a matrix of strings. */
    public String[][] getValueMatrix()
    {
	String[][] valueMatrix = new String[results.size()][names.length];
	for (int i = 0; i < results.size(); i++)
	{
	    valueMatrix[i] = getValues(i);
	}
	return valueMatrix;
    }

  /** Returns column index of named variable, or -1 if error--
    eg, name not there. */
    public int getNameIndex (String varName)
    {
	for (int i = 0; i < names.length; i++)
	{
	    if (names[i].equals(varName))
	    {
		return i;
	    }
	}

	return -1;
    }


  /** Set the name of a variable in a specific position. */
    public void setName(int pos, String n)
    {
	names[pos] = n;
    }

  /** Set the names of all variables. */
    public void setNames(String[] nArray)
    {
	names = nArray;
    }


/*
    public void setValue(int row, int col, String v)
    {
	values[row][col] = v;
    }

    public void setValues(int pos, String[] vArray)
    {
	values[pos] = vArray;
    }
*/


  /** Set the results based on the input matrix. */
    public void setValueMatrix (String[][] vMat)
    {
	if (vMat == null)
	{
	    return;
	}
	for (int i = 0; i < vMat.length; i++)
	{
	    results.addElement(new QueryResult(vMat[i]));
	}
    }


  /** Set the result based on the array of strings, where the rows
    have been linearized one after the other. This input form is the
    way Parka returns answers. */
    public void setValueMatrix (String[] answerList)
    {
	int  colSize = names.length;
	int  rowNum  = numResults();
	int  k = 0;

	if (answerList.length != colSize * rowNum)  //we have an error
	    System.out.println ("Answer list does not fit QueryResults matrix");
	if (answerList == null)
	    System.out.println ("Answer list is null");

	for (int i = 0; i < rowNum; i++)
	{
	    String[]  vals = new String[colSize];
	    for (int j = 0; j < colSize; j++)
	    {
		vals[j] = answerList[k];
		k++;
	    }
	    results.addElement (new QueryResult (vals));
	}
    }


  /** Remove duplicate answers(rows) from the results. */
    public boolean removeDuplicates()
    {
	boolean modified = false;

	// Check up to the second to last element...
	for (int i = 0; i < results.size() - 1; i++)
	{
	    // Check to see if it appears anywhere later in the vector...
	    QueryResult qr = (QueryResult) results.elementAt(i);
	    int j = i + 1;
	    j = results.indexOf(qr, i + 1);
	    while ((j != -1) && (j < results.size()))
	    {
		results.removeElementAt(j);     // remove duplicate
		// We could have deleted the last element...
		if (j < results.size())
		{
		    j = results.indexOf(qr, j);     // check for next possible
		}
	    }
	}
	results.trimToSize();     // shrink capacity

	return modified;
    }


///// old interfaces...
/*
    public int getNumResults()
    {
	return numResults();
    }

    public int getNumNames()
    {
	return numNames();
    }
*/

  /** A debugging tool to print out the current results as a table. */
    public String toString()
    {
	StringBuffer buff = new StringBuffer();
	for (int k = 0; k < names.length; k++)
	{
	    if ((k > 0) && (k % 8 == 0))
		buff.append ('\n');
	    buff.append (names[k] + " ");
	}
	buff.append("\n");
	buff.append(results);
	return   buff.toString();
    }
}


/** A single set of variable bindings. Represents a single answer from
  the KB. */
class QueryResult
{
    String[] valArray;

    QueryResult (String[] values)
    {
	valArray = values;
    }

  /** Two results are equal if they have equivalent bindings for each
    variable. */
    public boolean equals(Object o)
    {
	if (o instanceof QueryResult)
	{
	    QueryResult qr = (QueryResult) o;
	    for (int i = 0; i < valArray.length; i++)
	    {
		if (valArray[i].equals(qr.valArray[i]) == false)
		{
		    return false;
		}
	    }
	    return true;
	}
	else
	{
	    return false;
	}
    }

    public String toString()
    {
	StringBuffer buffer = new StringBuffer();
	if (valArray.length > 0)
	{
	    buffer.append(valArray[0]);
	}
	for (int i = 1; i < valArray.length; i++)
	{
	    buffer.append(" " + valArray[i]);
	}
	return buffer.toString();
    }
}
