/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ShoeKb.java
 * and the file License.html for details.
 */

package ShoeKb;

import java.io.*;
import java.util.Vector;
import SHOE.Def_Relation;

/** Implements the KBInterface by creating Parka data files which can
  then be used to create a new Parka knowledge base. */
public class ParkaFileInterface extends KBInterface {

  Vector      predVec;                   // contains a list of predicates
  PrintWriter shoePrint;
  PrintWriter predPrint;

  public ParkaFileInterface() {

  }

  /* Knowledge Base Interface Functions */

  /** Establish a connection to the given host and port number. */
  public void establishConnection(String newHost, int newPort)
       throws KBException
  {
    // No connection needs to be established
  }

  /** Creates a new KB with the name supplied by the user. */
  public void createKB(String kbName)
      throws KBException
  {
    String shoeFile;
    String predFile;
    FileOutputStream shoeStream;
    FileOutputStream predStream;
    Predicate pred;

    shoeFile = kbName + ".kb";
    try {
      shoeStream = new FileOutputStream(shoeFile);
      shoePrint =  new PrintWriter(shoeStream);
    } catch (IOException ex) {
      displayMessage("Couldn't open output file " + shoeFile);
      return;
    } 

    predFile = kbName + ".pred";
    try {
      predStream = new FileOutputStream(predFile);
      predPrint =  new PrintWriter(predStream);
    } catch (IOException ex) {
      displayMessage("Couldn't open output file " + predFile);
      return;
    } 

    /* create vector for storing predicates and insert an initial
       claims() predicate into it. */
    predVec = new Vector();
    pred = new Predicate("claims", 2);
    predVec.addElement(pred);
  }

  /** Opens an existing knowledge base using the default location. Parka
      files are only used to create new knowledge bases, so this method
      is unimplemented. */
  public void openKB(String kbName)
       throws KBException 
  {

  }

  /** Opens an existing knowledge base that is avaiable through the
    given host and port number. Parka files are only used to create
    new knowledge bases, so this method is unimplemented. */
  public void openKB(String kbName, String host, int port)
       throws KBException
    {
      
    }

  public void closeKB() {

    /* shoePrint will be uninitialized if no search was performed */
    if (shoePrint != null)
      shoePrint.close();
    
    // write out the entire predicate file
    writePredicates();      
    if (predPrint != null)
      predPrint.close();
  }

  /** Closes the current connection. Any open KBs will be closed as well. */
  public void closeConnection() {

    // Do nothing. This interfaces uses a local connection.
  }

  /*
  public ResultSet queryKB(String predicate, String args[]) {

    // cannot query the file being created to initialize a PARKA KB 
    return null;
  }
  */

  /*
  public ResultSet queryKB(String predicate, String arg1, String arg2) {

    // cannot query the file being created to initialize a PARKA KB
    return null;
  }
  */

  /** Issues an arbitrary query to the KB. Variable arguments are 
    preceded by question marks. The answers are returned in a matrix. */
  public ResultSet queryKB(Query query) {

    /* cannot query the file being created to initialize a PARKA KB */
    return null;
  }

  /* Creates a Parka claim given a source, predicate and array of 
     arguments. The arguments have escape codes inserted if they contain
     any special Parka characters. */
  public void storeClaim(String source, String predicate, String args[]) {

    String assert;
    String claim;
    int i;

    if (predicate == null)        // do nothing if no predicate is given
      return;
    assert = ParkaInterface.formatAssert(predicate, args);
    shoePrint.println(assert);
    // remove partentheses from claim
    claim = assert.substring(1, assert.length()-1);
    assert = ParkaInterface.formatAssert("claims", source, claim);
    shoePrint.println(assert);
  }

  /* A binary version of storeClaim(String, String, String[]) */
  /*
  public void storeClaim(String source, String predicate,
			 String arg1, String arg2) {

    String args[];

    args = new String[2];
    args[0] = arg1;
    args[1] = arg2;
    storeClaim(source, predicate, args);
  }
  */

  public void deleteClaims(String source) {

    /* cannot delete a claim from the file being created to initialize 
       a PARKA KB */
  }


  /** Asserts a new axiom to the KB. */
  public void storeAxiom(Axiom anAxiom) {

  }

  public void displayMessage(String msg) {

    System.out.println(msg);
  }

  /** Creates a new class with the specified name. */
  public void createClass(String className) {

  }

  /* Since Parka requires that the number of predicates in the file
     appear at the top of the file, we cannot write any predicates
     until we are done with the search. Therefore we save each
     predicate in a vector. */
  public void createPredicate(String predicate, int arity) {

    Predicate pred;

    pred = new Predicate(predicate, arity);
    predVec.addElement(pred);
  }

  /** Creates a new predicate with given name and types for
      arguments. */
  public void createPredicate(String predicate, String argTypes[]) {

    createPredicate(predicate, argTypes.length);
  }

  /* When creating a Parka file, you must have a count of the predicates
     at the top, followed by the list of the predicates. For this reason,
     this function takes an entire list of relations, counts them and
     adds them to a file that consists only of predicate definitions. */
  public void writePredicates() {

    int predCnt;
    int i;
    Predicate pred;

    predCnt = predVec.size();
    predPrint.println(predCnt);

    for (i=0; i < predCnt; i++) {
      pred = (Predicate)predVec.elementAt(i);
      predPrint.println(pred.predName + " " +pred. arity + " 0");
    }
  }

}

/** A simply data structure to store predicate information. */
class Predicate {

  String predName;
  int    arity;

  public Predicate(String predName, int arity) {

    this.predName = predName;
    this.arity = arity;
  }
}
